from ._misc import rk, rv
from ._misc import _check_and_format_integer_list, _check_integer_single, _check_dict_object

def run_trials_miniBatchKMeans(data, n_clusters, d_range, n_runs, batch_size, random_state):
    """
    Generate dictionary object containing MiniBatchKMeans clustering runs.
    """

    import itertools
    from sklearn.cluster import MiniBatchKMeans

    # check that consensus matrix has no NaN, infinity values
    from sklearn.utils import check_array
    data = check_array(data)

    # check n_clusters
    n_clusters = _check_integer_single(
        n_clusters,
        min_val = 2,
        var_name = 'n_clusters'
    )

    # check d_range
    d_range = _check_and_format_integer_list(
        d_range,
        min_val = 2,
        max_val = data.shape[1],
        var_name = 'd_range'
    )

    # check n_runs
    n_runs = _check_integer_single(n_runs, min_val=1, var_name="n_runs")

    # check batch_size
    batch_size = _check_integer_single(
        batch_size,
        min_val = 10,
        max_val = data.shape[0],
        var_name = 'batch_size'
    )

    # check random state
    from sklearn.utils import check_random_state
    random_state = check_random_state(random_state)

    trials_dict = {}
    for d, i in itertools.product(d_range, range(1, n_runs + 1)):
        kmeans = MiniBatchKMeans(
            n_clusters = n_clusters,
            batch_size = batch_size,
            random_state = random_state
        )

        # cluster the points
        kmeans.fit(data[:, :d])

        # add results into dictionary
        # trials_dict[rk(d, i)] = rv(  # NEED SCANPY TO IMPLEMENT SAVING
        trials_dict[(d, i)] = rv(
            facility = kmeans.cluster_centers_,
            labels = kmeans.labels_,
            inertia = kmeans.inertia_
        )
    
    return trials_dict


def combine_facilities(dict_object, K, n_facility, batch_size, random_state):
    """
    Combine facilities in a dictionary object.
    """
    from sklearn.cluster import MiniBatchKMeans
    import numpy as np

    # check that the dictionary object is formatted correctly
    assert isinstance(dict_object, dict)
    for key, value in dict_object.items():
        # assert isinstance(key, rk)  # NEED SCANPY TO IMPLEMENT SAVING
        assert isinstance(value, dict)
        assert value['facility'] is not None
        assert value['labels'] is not None

    # check arguments are formated as integers (limits not trivial to do)
    K = _check_integer_single(K, var_name = 'K')
    n_facility = _check_integer_single(n_facility, var_name = 'n_facility')
    batch_size = _check_integer_single(batch_size, var_name = 'batch_size')

    # check random state
    from sklearn.utils import check_random_state
    random_state = check_random_state(random_state)

    dict_object_combined = {}

    # combine the facilities from different trials
    for key, value in dict_object.items():
        facilities = value['facility']
        labels = value['labels']

        # count the number of cells assigned to each facility
        weights = list(map(
            lambda cluster_no: np.count_nonzero(labels == cluster_no), 
            range(0, n_facility)
        ))

        # cluster the facility using weighted k means
        kmeans = MiniBatchKMeans(
            n_clusters = K,
            batch_size = batch_size,
            random_state = random_state
        )
        kmeans.fit(facilities, sample_weight = weights)

        # write result
        dict_object_combined[key] = rv(
            facility = kmeans.cluster_centers_,
            labels = kmeans.labels_[labels], 
            inertia = kmeans.inertia_
        )

    return dict_object_combined



def convert_dict_into_binary_matrix(dict_object, true_n_clusters, true_n_cells):
    """ 
    Convert clustering runs in a dictionary into a binary consensus matrix.
    """

    import numpy as np
    import logging

    n_clusters, n_cells = true_n_clusters, true_n_cells

    # check dict_object has correct number of clusters and cells for every run
    dict_object = _check_dict_object(dict_object, true_n_clusters, true_n_cells)

    # initialise empty B array with the correct shape
    B = np.zeros((n_cells, 0), dtype=int)

    # for each run, we create the block to be appended to B
    for i, value in enumerate(dict_object.values()):
        cell_labels = value['labels']

        # create a total of 'n_clusters' columns for this iteration
        b = np.zeros((n_cells, n_clusters), dtype=int)

        # annotate the correct columns for each row/cell
        b[range(0, n_cells), cell_labels] = 1

        # this checks that every cell only has one cluster assignment
        assert np.all(np.sum(b, axis=1) == 1), "some cells have multiple cluster assignments"

        # append to the right of the overall binary consensus matrix
        B = np.append(B, b, axis=1)

    # remove columns in B with no cells assigned 
    # this happens if some initialised centroids do not get assigned anything
    logging.info(f'original binary matrix shape: {B.shape}')
    cols_with_no_cells = np.sum(B, axis=0) != 0
    if not np.all(cols_with_no_cells):
        B = B[:, np.sum(B, axis=0) != 0]
        logging.warning(f'remove some clusters with no assigned cells, binary matrix shape is now: {B.shape}')  

    return B


def cluster_consensus_matrix(consensus_matrix, n_clusters, batch_size, random_state):
    """
    Cluster the consensus binary matrix using miniBatchKmeans. Rows of matrix should be observations.
    """

    # check that consensus matrix has no NaN, infinity values
    from sklearn.utils import check_array
    consensus_matrix = check_array(consensus_matrix)

    # check n_clusters
    n_clusters = _check_integer_single(
        n_clusters,
        min_val = 2,
        max_val = consensus_matrix.shape[0],
        var_name = 'n_clusters'
    )

    # check batch_size
    batch_size = _check_integer_single(
        batch_size,
        min_val = 10,
        max_val = consensus_matrix.shape[0],
        var_name = 'batch_size'
    )

    # check random state
    from sklearn.utils import check_random_state
    random_state = check_random_state(random_state)

    from sklearn.cluster import MiniBatchKMeans
    kmeans = MiniBatchKMeans(
        n_clusters = n_clusters,
        batch_size = batch_size,
        random_state = random_state
    )
    kmeans.fit(consensus_matrix)

    return kmeans.labels_


from ._misc import _check_integer_single, _check_dict_object

def convert_dict_into_contigency_matrix(dict_object, true_n_clusters, true_n_cells):
    """
    Convert clustering runs in a dictionary into a contingency consensus matrix.
    """
    
    import numpy as np

    n_clusters, n_cells = true_n_clusters, true_n_cells
    n_trials = len(dict_object.values())

    # check dict_object has correct number of clusters and cells for every run
    dict_object = _check_dict_object(dict_object, true_n_clusters, true_n_cells)

    # initialise contingeny matrix
    C = np.zeros((n_cells, n_cells))

    # obtain the clustering labels from each trial as a matrix
    cell_labels_matrix = np.array([trial['labels'] for trial in dict_object.values()]).T
    assert cell_labels_matrix.shape == (n_cells, n_trials)

    # count the number of matches for each cell
    for cell in range(0, n_cells):
        C[cell, ] = np.sum(cell_labels_matrix[cell, :] == cell_labels_matrix, axis = 1) / n_trials

    # this might be expensive
    assert np.allclose(C, C.T), "something went wrong, contingency matrix not symmetrical"
    
    return C


def cluster_matrix_kmeans(consensus_matrix, n_clusters, random_state):
    """
    Wrapper around scikit learn whole batch k means. Rows of matrix should be observations.
    """

    # check that consensus matrix has no NaN, infinity values
    from sklearn.utils import check_array
    consensus_matrix = check_array(consensus_matrix)

    # check n_clusters
    n_clusters = _check_integer_single(
        n_clusters,
        min_val = 2,
        max_val = consensus_matrix.shape[0],
        var_name = 'n_clusters'
    )

    # check random state
    from sklearn.utils import check_random_state
    random_state = check_random_state(random_state)

    from sklearn.cluster import KMeans
    kmeans = KMeans(n_clusters=n_clusters, random_state=random_state)
    kmeans.fit(consensus_matrix)

    return kmeans.labels_

def cluster_matrix_agglomerative(consensus_matrix, n_clusters):
    """
    Wrapper around scikit learn agglomerative clustering. Rows of matrix should be observations.
    """

    # check that consensus matrix has no NaN, infinity values
    from sklearn.utils import check_array
    consensus_matrix = check_array(consensus_matrix)

    # check n_clusters
    n_clusters = _check_integer_single(
        n_clusters,
        min_val = 2,
        max_val = consensus_matrix.shape[0],
        var_name = 'n_clusters'
    )

    from sklearn.cluster import AgglomerativeClustering
    agg_clust = AgglomerativeClustering(n_clusters=n_clusters)
    agg_clust.fit(consensus_matrix)

    return agg_clust.labels_


import pandas
from sklearn.utils import check_random_state

from ._cluster import run_trials_miniBatchKMeans
from ._cluster import combine_facilities
from ._cluster import convert_dict_into_binary_matrix, cluster_consensus_matrix

from ._misc import _check_and_format_integer_list, _check_integer_single

import logging

def consensus(
    adata,
    n_clusters = [3, 5, 10],
    d_range = None,
    n_runs = 5,
    n_facility = None,
    multiplier_facility = None,
    batch_size = None,
    random_state = None):
    """\
    Run consensus clustering to cluster cells in an AnnData object.
    
    This function requires the user to perform dimensionality
    reduction using PCA (`scanpy.tl.pca`) first.

    Parameters
    ----------
    adata
        Annotated data matrix.
    n_clusters
        Number of clusters. Default: [3,5,10]
    d_range
        Number of PCs. Default is 25, or the number of PCs in the 
        AnnData object, whichever is lower. Can accept a list
        (e.g. `[15, 20, 25`]).
    n_runs
        Number of realisations to perform for the consensus.
        Default is 5, recommended > 1.
    n_facility
        Number of microclusters. Overridden if `multiplier_facility`
        is provided with a value.
    multiplier_facility
        Multiplier for microclusters. Number of microclusters
        is calculated as this parameter multiplied by the max of
        `n_clusters`. Default is 3.
    batch_size
        Batch size for k-means. Default is 100.
    random_state
        Random state of the algorithm.

    Returns
    -------
    AnnData object with cluster labels written in the `.obs` dataframe.

    adata.obs['sc3s_{k}']
        Labels for `n_clusters = k`.
    """

    logging.basicConfig(level=logging.INFO)

    # check that AnnData object already has PCA coordinates
    try:
        X_pca = adata.obsm['X_pca']
    except:
        raise Exception("AnnData object does not have PCA coordinates. Please run that first.")

    # get the number of cells from row dimension
    n_cells = X_pca.shape[0]

    # check and formats n_clusters into in a list format, using default values if unprovided
    n_clusters = _check_and_format_integer_list(
        n_clusters, 
        min_val = 2, 
        max_val = n_cells,
        var_name = 'n_clusters'
    )

    # check d_range
    if d_range is None:
        d_range = [min(25, X_pca.shape[1])]
    d_range = _check_and_format_integer_list(
        d_range,
        min_val = 2,
        max_val = X_pca.shape[1],
        var_name = 'd_range'
    )

    # check n_runs
    n_runs = _check_integer_single(n_runs, min_val=1, var_name="n_runs")

    # check batch_size
    if batch_size is None:
        batch_size=min(100, adata.shape[0])
    if isinstance(batch_size, int):
        assert batch_size <= n_cells, "batch size for k-means must be smaller than number of cells."
    else:
        raise Exception("batch size must be positive integer value (default: 100)")

    # check random state
    random_state = check_random_state(random_state)
 
    # determine number of facilities to use
    if n_facility is not None:
        logging.info(f"n_facility set to {n_facility}, ignoring multiplier_facility parameter...")
    else:
        if multiplier_facility is None:
            logging.info(f"multiplier_facility not set, using default value of 3...")
            multiplier_facility = 3
        
        # use multiplier value to calculate
        n_facility = max(n_clusters) * multiplier_facility

        if n_facility >= n_cells:
            n_facility = n_cells - 1

        logging.info(f"number of facilities calculated as {n_facility}")    
    
    n_facility = _check_integer_single(
        n_facility, 
        min_val = 2, 
        max_val = n_cells,
        var_name = 'n_facility'
    )

    # run over many different trials
    trials_dict = run_trials_miniBatchKMeans(
        data = X_pca,
        n_clusters = n_facility,
        d_range = d_range,
        n_runs = n_runs,
        batch_size = batch_size,
        random_state = random_state
    )

    # write the individual trials into AnnData object
    adata.uns['sc3s_trials'] = trials_dict

    # run different number of clusters
    for K in n_clusters:
        consensus_dict = combine_facilities(
            dict_object = trials_dict, 
            K = K, 
            n_facility = n_facility,
            batch_size = batch_size, 
            random_state = random_state
        )

        B = convert_dict_into_binary_matrix(
            dict_object = consensus_dict,
            true_n_clusters = K,
            true_n_cells = n_cells
        )

        cell_labels = cluster_consensus_matrix(
            consensus_matrix = B,
            n_clusters = K,
            batch_size = batch_size,
            random_state = random_state
        )

        # write results into AnnData object
        adata.obs[f'sc3s_{K}'] = cell_labels
        adata.obs[f'sc3s_{K}'] = adata.obs[f'sc3s_{K}'].astype('category')

    return adata


# special objects for the trials dictionary 
from collections import namedtuple

# key
rk = namedtuple('trial_params', 'd, i')

# value
def rv(**kwargs):
    """
    Basically a nice wrapper around kwargs to avoid having to type out dictionary syntax.

    Returns dictionary object.
    """
    return kwargs

def _check_and_format_integer_list(num_list, min_val=None, max_val=None, var_name="variable", return_as_list = True):
    """
    Format a list/range/single integer into iterable format.
    
    Also checks that values fall within the limits specified.
    """
    # formats into an iterable
    _check_integer_boolean = isinstance(num_list, int)

    if _check_integer_boolean:
        num_list = [num_list]
    elif isinstance(num_list, range):
        if len(num_list) == 0:
            raise Exception(f"{var_name} specified range has length 0.")
        else:
            pass
    elif isinstance(num_list, list):
        if len(num_list) == 0:
            raise Exception(f"{var_name} specified list has length 0.")
        else:
            for num in num_list:
                if not isinstance(num, int):
                    raise Exception(f"{var_name} values must be integers.")
                else:
                    pass
    else:
        raise Exception(f"{var_name} must be integer, or a non-empty list/range.")

    # check that each value in the iterable satisfies the limits
    for num in num_list:
        if min_val is not None:
            if num < min_val:
                raise Exception(f"{var_name} must be greater than {min_val}.")
        if max_val is not None:
            if num > max_val:
                raise Exception(f"{var_name} must be less than {max_val}.")

    # the formated integer can be returned in the original format if specified
    if _check_integer_boolean and return_as_list == False:
        num_list = num_list[0]
    
    return num_list


def _check_integer_single(num, min_val=None, max_val=None, var_name="variable"):
    """
    Checks that a value is an integer.
    
    Also checks that the value fall within the limits specified.
    """

    # check that it is an integer
    if not isinstance(num, int):
        raise Exception(f"{var_name}: must be an integer.")

    # check that the integer value satisfies the limits
    if min_val is not None:
        if num < min_val:
            raise Exception(f"{var_name} must be greater than {min_val}.")
    if max_val is not None:
        if num > max_val:
            raise Exception(f"{var_name} must be less than {max_val}.")

    return num

def _check_dict_object(dict_object, true_n_clusters, true_n_cells):
    """
    Check that the runs in the dict_object contain the specified number of n_clusters and n_cells.
    """

    # check n_clusters and n_cells are integers
    true_n_clusters = _check_integer_single(true_n_clusters)
    true_n_cells    = _check_integer_single(true_n_cells)

    # check the number of cells
    n_cells_list = [len(x['labels']) for x in dict_object.values()]
    assert n_cells_list.count(true_n_cells) == len(n_cells_list), "number of cells is not consistent between trials"
    n_cells = true_n_cells

    # check the number of clusters
    n_clusters_list = [x['facility'].shape[0] for x in dict_object.values()]
    assert n_clusters_list.count(true_n_clusters) == len(n_clusters_list), "number of clusters not consistent between trials"
    n_clusters =  true_n_clusters

    return dict_object


    # check the number of cells
    # n_cells_list = [len(x['labels']) for x in dict_object.values()]
    # assert n_cells_list.count(true_n_cells) == len(n_cells_list), "number of cells not consistent"
    # n_cells = true_n_cells

    # # check the number of clusters
    # # there is an edge case for this, which is if the kth cluster is not assigned any cells
    # # will fix if error occurs, otherwise can just count that at least 75% of runs is correct?
    # n_clusters_list = [np.max(x['labels']) + 1 for x in dict_object.values()]
    # assert n_clusters_list.count(true_n_clusters) == len(n_clusters_list), "number of cells not consistent"
    # n_clusters =  true_n_clusters



from setuptools import setup, find_packages

setup(
    name="sc3s", 
    version="0.1.1",
    description="Python Single Cell Consensus Clustering",
    url="https://github.com/pittachalk/sc3s",
    author="F.X. Quah",
    author_email="fxq20@cam.ac.uk",
    classifiers=[
       "Programming Language :: Python :: 3",
       "License :: OSI Approved :: BSD License",
       "Operating System :: OS Independent"
    ],
    packages=find_packages(),
    python_requires=">=3.6",    
    install_requires=[
       "numpy>=1.19.2,<2",
       "pandas>=1.1.5",
       "scikit-learn>=0.23.2",
       "scanpy>=1.7.0",
       "six>=1.15.0",
       "setuptools>=52.0.0",
       "joblib>=1.0.0",
       "h5py >=3.1.0",
       "seaborn",
       "scipy",
       "python-dateutil >=2.8.1",
  ]
)
